# MIT License
#
# Copyright (c) 2017 Evan Liu (hmisty)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
enum
"""
def enum(**enums):
    return type('Enum', (), enums)

connection_mode = enum(LOCAL = 'LOCAL',
                       STAND_ALONE = 'STAND_ALONE',
                       STAND_BY = 'STAND_BY',
                       LOAD_BALANCE = 'LOAD_BALANCE')

"""
from common import multimethod

@multimethod(int, int)
def foo(a, b):
    ...

@multimethod(int)
def foo(a):
    return foo(a, 10) # This calls the previous foo()!
"""
registry = {}

class MultiMethod(object):
    def __init__(self, name):
        self.name = name
        self.typemap = {}

    def __call__(self, *args):
        types = tuple(arg.__class__ for arg in args) # a generator expression!
        print types
        
        function = self.typemap.get(types)
        if function is None:
            raise TypeError("no match")
        return function(*args)

    def register(self, types, function):
        if types in self.typemap:
            raise TypeError("duplicate registration")
        self.typemap[types] = function

def multimethod(*types):
    def register(function):
        name = function.__name__
        mm = registry.get(name)
        if mm is None:
            mm = registry[name] = MultiMethod(name)
        mm.register(types, function)
        return mm

    return register




# MIT License
#
# Copyright (c) 2017 Evan Liu (hmisty@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import print_function
import os
import sys
import signal
import atexit
import time
import socket
import select
import psutil
import traceback

from .config import settings

#log
def log(*args, **kwargs):
    pid = os.getpid()
    print(pid, *args, **kwargs)

# get daemon pid
def get_daemon_pid():
    try:
        pf = file(settings.pid_file, 'r')
        pid = int(pf.read().strip())
        pf.close()
    except IOError:
        pid = None
    except SystemExit:
        pid = None

    return pid

# delete daemon pid file
def del_daemon_pid():
    if os.path.exists(settings.pid_file):
        log('delete daemon pid file %s' % settings.pid_file)
        os.remove(settings.pid_file)

# fork a worker
def start_worker(worker_main_loop):
    pid = os.fork()

    if pid:
        # in parent process
        return pid
    else:
        # fork pid == 0, in child process

        # clear signal hooks
        signal.signal(signal.SIGCHLD, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.signal(signal.SIGINT, signal.SIG_DFL)

        # start loop
        try:
            worker_main_loop()
        except Exception, e:
            log('worker main loop exit', repr(e))
            sys.exit(1)

# daemonize
def daemonize():
    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)
    except OSError, e:
        log('fork1 failed', repr(e))
        sys.exit(1)

    os.chdir(settings.home_dir)
    os.setsid()
    os.umask(settings.umask)

    try:
        pid = os.fork()
        if pid > 0:
            sys.exit(0)
    except OSError, e:
        log('fork2 failed', repr(e))
        sys.exit(1)

    sys.stdout.flush()
    sys.stderr.flush()

    si = file(os.devnull, 'r')
    so = file(settings.log_file, 'a+', 0)
    se = file(settings.err_file, 'a+', 0)

    os.dup2(si.fileno(), sys.stdin.fileno())
    os.dup2(so.fileno(), sys.stdout.fileno())
    os.dup2(se.fileno(), sys.stderr.fileno())

    log('daemon process started')

    #atexit.register(del_daemon_pid) #buggy. child will inherit and del pid file when die
    pid = str(os.getpid())
    file(settings.pid_file, 'w+', 0).write('%s\n' % pid)

    log('daemon pid_file(%s) created' % settings.pid_file)

# quit
def quit_daemon():
    # daemon entering exiting process
    log('stopping all workers ...')

    # kill all the workers
    signal.signal(signal.SIGCHLD, signal.SIG_IGN)

    pid = os.getpid()
    worker_list = psutil.Process(pid).children()
    for p in worker_list:
        try:
            log('killing %s ...' % p.pid)
            p.kill()
        except OSError, err:
            err = repr(err)
            log(err)
            pass

    del_daemon_pid()

    log('daemon %s exit' % pid)
    sys.exit(0)

# init
def init_daemon():
    # daemonize the parent
    # * status workers
    # * start/stop/restart a worker
    # * auto-restart if a worker dies
    pid = get_daemon_pid()
    if pid:
        log('%s already running' % pid)
        sys.exit(1)
    else:
        daemonize()

# exported
def setup(local_settings):
    settings.update(local_settings)

# exported
def start(worker_main_loop, n_workers=None):
    log('daemon starting...')
    init_daemon()

    if n_workers:
        settings.n_workers = n_workers

    workers = []
    for i in range(settings.n_workers):
        pid = start_worker(worker_main_loop)
        workers.append(pid)

    log('workers', workers, 'started running!')

    # auto-restart child
    def child_exit_handler(signum, frame):
        pid = os.wait()[0]
        log('worker(%s) died. fork a new...' % pid)
        time.sleep(1.0)
        start_worker(worker_main_loop)

    signal.signal(signal.SIGCHLD, child_exit_handler)

    # respond to termination
    def terminate_handler(signum, frame):
        log('daemon being terminated...')
        quit_daemon()

    signal.signal(signal.SIGTERM, terminate_handler)

    def interrupt_handler(signum, frame):
        log('daemon being interrupted...')
        quit_daemon()

    signal.signal(signal.SIGINT, interrupt_handler)

    # forever loop
    log('daemon', os.getpid(), 'started running!')
    while True:
        time.sleep(0.1)

"""
exported
"""
def stop():
    log('daemon stopping...')

    pid = get_daemon_pid()
    if not pid:
        pid_file = settings.pid_file
        msg = 'pid file [%s] does not exist. is it not running?' % pid_file
        log(msg)
        if os.path.exists(pid_file):
            os.remove(pid_file)

        return

    #try to kill the daemon process
    try:
        os.kill(pid, signal.SIGTERM)
    except OSError, err:
        err = repr(err)
        log(err)
        if err.find('No such process') > 0:
            pid_file = settings.pid_file
            if os.path.exists(pid_file):
                os.remove(pid_file)
            else:
                log(repr(err))
                sys.exit(1)


# exported
def status():
    pid = get_daemon_pid()
    if not pid:
        pid_file = settings.pid_file
        msg = 'pid file [%s] does not exist. is it not running?' % pid_file
        print(msg)
        if os.path.exists(pid_file):
            os.remove(pid_file)

        return

    print('status: running')
    print('%s: daemon' % pid)

    worker_list = psutil.Process(pid).children()
    for p in worker_list:
        if psutil.pid_exists(p.pid):
            print('%s: active worker' % p.pid)
        else:
            print('%s: dead worker' % p.pid)




# MIT License
#
# Copyright (c) 2017 Evan Liu (hmisty@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


class Settings:
    _ = {
        'host': '127.0.0.1',
        'port': 8181,
        'n_workers': 2,
        'home_dir': '/tmp',
        'umask': 022,
        'pid_file': '/tmp/bson-rpc.pid',
        'log_file': '/tmp/bson-rpc.out',
        'err_file': '/tmp/bson-rpc.err',
        'host_file': '/tmp/bson-rpc.host', # which interface to bind
    }

    def __repr__(self):
        return repr(self._)

    def __getattr__(self, name):
        return self._[name]

    def __setattr__(self, name, value):
        self._[name] = value

    def update(self, settings):
        self._.update(settings)

# use the instance
settings = Settings()




# MIT License
#
# Copyright (c) 2017 Evan Liu (hmisty@gmail.com)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import print_function
import os
import sys
import signal
import atexit
import time
import socket
import select
import Queue
import bson

import status

# the global function map that remotely callable
# { fn: [func, invoke_count, accumulated_time] }
remote_functions = dict()

def rpc(func, name=None):
    """ add a function to remote callable function map.

    >>> from bson_rpc import rpc

    use as function

    >>> rpc(lambda s: s, name="echo")

    or use as decorator

    >>> @rpc
    def echo(s):
        return s

    """
    global remote_functions
    remote_functions[name or func.__name__] = [func, 0, 0]
    return func

def start(host = '127.0.0.1', port = 8181):
    """
    start_server

    >>> from bson_rpc import server
    >>> server.start()
    """
    server = Server(host, port)

    print("server started, listening on %s:%s" % (host, port))
    server.start_forever()

@rpc
def __stats__():
    rf = remote_functions
    stats = {}
    total_count = 0
    total_time = 0
    for fn in remote_functions:
        count = remote_functions[fn][1]
        time = remote_functions[fn][2]
        stats[fn] = [count, time]
        total_count += count
        total_time += time

    stats['*'] = [total_count, total_time]

    return stats

def invoke_func(fn, args=None):
    global remote_functions

    if fn is None:
        return status.function_not_found.copy()

    if not remote_functions.has_key(fn):
        return status.function_not_found.copy()

    f = remote_functions[fn][0]
    if not callable(f):
        return status.function_not_callable.copy()

    try:
        begin = time.time()

        if args == None:
            result = f()
        else:
            result = f(*args)

        elapsed = time.time() - begin
        elapsed = int(elapsed * 1000) # in milliseconds

        response = status.ok.copy() # copy() before modify!
        response['result'] = result
        response['time'] = elapsed

        remote_functions[fn][1] += 1
        remote_functions[fn][2] += elapsed
    except Exception as error:
        response = status.invoke_error.copy()
        response['error_msg'] = repr(error)

    return response

def compute_on(obj):
    response = None

    # obj is a bson obj received from a socket
    if obj != None:
        _id = obj.get('_id') or 0

        fn = obj.get('fn')
        args = obj.get('args')
        response = invoke_func(fn, args)

        response['_id'] = _id

    return response

def log(sock, *args):
    datetime = '%s %s' % (time.ctime(), time.tzname[0])

    try:
        conn_str = '%s:%s' % sock.getpeername()
    except:
        conn_str = 'unconnected'

    message = ' '.join(map(lambda x: str(x), args)).replace('\n', '').replace('\r', '')
    print(datetime, conn_str, message.decode('unicode_escape'))

class Server:
    def __init__(self, host, port):
        self.pid = 0 # pid of the worker process
        self.host = host
        self.port = port

        server = self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server.setblocking(False)
        server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server.bind((self.host, self.port))
        server.listen(5) # allow max 5 in waiting list
        log(server, 'Listening on', self.host, self.port)

        self.inputs = [self.server] # sockets to read
        self.outputs = [] # sockets to write
        self.message_queues = {} # socket message queue

    def start_forever(self, polling_interval=60):
        bson.patch_socket()

        server = self.server
        inputs = self.inputs
        outputs = self.outputs
        message_queues = self.message_queues

        inputs = self.inputs
        while inputs:
            readable, writable, exceptional = \
                    select.select(inputs, outputs, inputs, polling_interval)

            if not (readable or writable or exceptional):
                # timeout will generate three empty lists
                continue; # loop forever

            self.read_each(readable)
            self.write_each(writable)
            self.catch_each(exceptional)

    def read_each(self, readable):
        server = self.server
        inputs = self.inputs
        outputs = self.outputs
        message_queues = self.message_queues

        for sock in readable:
            if sock is server:
                # it is the server socket

                # while trying to accept, we need to handle the "thundering bird" problem
                # because in *nix, all child processes will be waken up to accept
                # only one child can succeed
                # others will fail to accept and go back to loop again
                # be careful not to crash when accept failed
                try:
                    conn, addr = sock.accept()
                except:
                    return

                log(conn, 'CONNECT')
                conn.setblocking(False)
                inputs.append(conn)
                message_queues[conn] = Queue.Queue()

            else:
                # it is a connection socket

                obj = None
                try:
                    obj = sock.recvobj()
                except Exception as e:
                    log(sock, e)

                if obj:
                    response = compute_on(obj) # caution: would block!
                    log(sock, 'REQUEST', obj)
                    message_queues[sock].put(response)

                    if sock not in outputs:
                        outputs.append(sock)
                    else:
                        pass

                else:
                    # treat empty message as closed connection
                    log(sock, 'DISCONNECT')
                    if sock in outputs:
                        outputs.remove(sock)

                    inputs.remove(sock)
                    sock.close()
                    del message_queues[sock]

    def write_each(self, writable):
        outputs = self.outputs
        message_queues = self.message_queues

        for sock in writable:
            try:
                if (sock in message_queues):
                    obj = message_queues[sock].get_nowait()
                    log(sock, 'REPLY', obj)
                    sock.sendobj(obj)
            except Queue.Empty:
                outputs.remove(sock)
            except socket.error, msg:
                log(sock, 'socket.error', msg)

    def catch_each(self, exceptional):
        inputs = self.inputs
        outputs = self.outputs
        message_queues = self.message_queues

        for sock in exceptional:
            log(sock, 'EXCEPTION')
            if sock in outputs:
                outputs.remove(sock)

            inputs.remove(sock)
            sock.close()
            del message_queues[sock]




__version = (0, 10, 3)

__version__ = version = '.'.join(map(str, __version))
__project__ = PROJECT = __name__

"""
>>> from bson_rpc import rpc, start_server
>>> from bson_rpc import connect
"""
#__all__ = ['rpc', 'start_server', 'connect']

import server
rpc = server.rpc
start_server = server.start

import client
connect = client.connect

import sys
defaultencoding = 'utf-8'
if sys.getdefaultencoding() != defaultencoding:
    reload(sys)
    sys.setdefaultencoding(defaultencoding)




#!/usr/bin/env python

# MIT License
#
# Copyright (c) 2017 Evan Liu (hmisty)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from distutils.core import setup

VERSION = "0.10.3"
URLBASE = "https://github.com/hmisty/bson-rpc/"
URLMAP = {
    "daily": "tarball/master",
    "1.1": "tarball/1.1",
}

if __name__ == "__main__":
    setup(
        name='bson-rpc',
        version=VERSION,
        description='a lightweight, high performance, multilingual RPC library',
        author='Evan Liu (hmisty)',
        author_email='hmisty@gmail.com',
        url=URLBASE,
        download_url='/'.join([URLBASE, URLMAP.get(VERSION, URLMAP['daily'])]),
        packages=[
            'bson_rpc'
        ],
        scripts=[],
        license= 'MIT',
        keywords = ['BSON', 'bson-rpc', 'brpc', 'rpc'],
        classifiers = [
            'Development Status :: 4 - Beta',
            'Topic :: Internet',
            'Environment :: No Input/Output (Daemon)',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 2.7',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        install_requires=['bson==0.4.8', 'psutil==5.4.1']
    )



# MIT License
#
# Copyright (c) 2017 Evan Liu (hmisty)
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from __future__ import print_function
from socket import socket
import bson
import os
import random

from common import connection_mode

class Proxy:

    # init some default setup
    def __init_defaults__(self):
        self.DIE_ON_FAILURE = True
        self.connection_mode = connection_mode.LOCAL
        self.hosts = ['127.0.0.1']
        self.port = 8181

        bson.patch_socket()

    # init
    def __init__(self, *args):
        self.__init_defaults__()

        if len(args) == 3: 
            # master, standby, port
            print('use STAND_BY mode')
            self.connection_mode = connection_mode.STAND_BY
            self.hosts = list(args[0:2])
            self.port = args[2]
        elif len(args) == 2 or len(args) == 1:
            if isinstance(args[0], str): # str == <type 'str'> 
                # host, port
                print('use STAND_ALONE mode')
                self.connection_mode = connection_mode.STAND_ALONE
                self.hosts = list(args[0:1])
                self.port = args[1] if len(args) > 1 else self.port
            elif isinstance(args[0], list): 
                # [host], port
                print('use LOAD_BALANCE mode')
                self.connection_mode = connection_mode.LOAD_BALANCE
                self.hosts = list(args[0])
                self.port = args[1] if len(args) > 1 else self.port
            else:
                print('incorrect type of host(s)')
                pass
        else:
            print('use LOCAL mode by default')
            pass

    # connect
    def connect(self):
        if self.connection_mode == connection_mode.LOCAL or self.connection_mode == connection_mode.STAND_ALONE:
            pair = (self.hosts[0], self.port)
            print('trying to connect %s:%s in %s mode' % (pair + (self.connection_mode,)))
            try:
                self.sock = socket()
                self.sock.settimeout(1) # fail asap after 1 sec
                self.sock.connect(pair)
                print('connected')
            except Exception, e:
                self.sock.close()
                self.sock = None
                self.fail(repr(e))

        elif self.connection_mode == connection_mode.STAND_BY:
            for host in self.hosts:
                pair = (host, self.port)
                print('trying to connect %s:%s in %s mode' % (pair + (self.connection_mode,)))
                try:
                    self.sock = socket()
                    self.sock.settimeout(1) # fail asap after 1 sec
                    self.sock.connect(pair)
                    print('connected')
                    break
                except Exception, e:
                    self.sock.close()
                    self.sock = None
                    print('failed: ' + repr(e))
                    pass
            
            if self.sock == None:
                self.fail('all failed')

        elif self.connection_mode == connection_mode.LOAD_BALANCE:
            hosts = self.hosts
            
            while len(hosts) > 0:
                host = random.choice(hosts)
                hosts.remove(host)
                pair = (host, self.port)
                print('trying to connect %s:%s in %s mode' % (pair + (self.connection_mode,)))
                try:
                    self.sock = socket()
                    self.sock.settimeout(1) # fail asap after 1 sec
                    self.sock.connect(pair)
                    print('connected')
                    break
                except Exception, e:
                    self.sock.close()
                    self.sock = None
                    print('failed: ' + repr(e))
                    pass
            
            if self.sock == None:
                self.fail('all failed')

    # use some services
    def use_service(self, namelist):
        namelist.append('__stats__')
        for name in namelist:
            f = self.invoke_func(name);
            setattr(self, name, f);

    # invoke a function
    def invoke_func(self, name):
        def invoke_remote_func(*args, **kwargs):
            try:
                self.sock.sendobj({'fn': name, 'args': list(args)})
                doc = self.sock.recvobj()
                err = doc['error_code']
            except Exception, e:
                self.sock.close()
                self.sock = None
                self.fail(repr(e))
            else:
                # reaches here if no exception
                if (err == 0):
                    result = doc['result']
                else:
                    result = doc['error_msg']

                return err, result

        return invoke_remote_func

    # client disconnect
    def disconnect(self):
        self.sock.close()
        self.sock = None
        self.fail('client disconnect explicitly')

    # Set DOF value, True by default
    def die_on_failure(self, dof=True):
        self.DIE_ON_FAILURE = dof

    # Fail the process
    def fail(self, error_msg = None):
        if self.DIE_ON_FAILURE:
            print('Die-On-Failure: ', error_msg)
            os._exit(1)
        else:
            raise Exception(error_msg)

# Helper function.
# Equivalent to,
# from bson_rpc import client
# proxy = client.Proxy(host, port)
def connect(*args):
    proxy = Proxy(*args)
    proxy.connect()
    return proxy

